#include"glnn.h"
#include<eigen3/Eigen/Dense>

//Version for tanh activation function
void GLNN::initWeights()
{
 int i,e;long y;
 long double s;
 long double alpha=.5;
 long double gamma[n];
 long double beta[n];
 long double read_epsilon;
 //Put the right global frequencies
 for(y=0;y<alph;y++)wbias[y]=logl(predfreq[y]);
 //for(y=0;y<alph;y++)wbias[y]=logl(predfreq[y])*(1.+1./30000.*sqrtl(predfreq[y])*alea_norm());//30000=data length
 //gamma must be strictly between alpha and 1
 for(i=0;i<n;i++)gamma[i]=1.-1./(i+2.);
 for(i=0;i<n;i++){
	 for(y=0;y<alph;y++)w[y][i]=0;
	 read_epsilon=(1.-gamma[i])/4.;
	 //alea() returns a uniform random value in [0;1]
	 for(y=0;y<alph;y++)taubias[y][i]=alea()*read_epsilon;
	 //centering
	 s=0;for(y=0;y<alph;y++)s+=taubias[y][i]*readfreq[y];
	 for(y=0;y<alph;y++)taubias[y][i]-=s;
	 //adding constant to get desired equilibrium value
	 beta[i]=-sqrtl(alpha*(gamma[i]+alpha-1));
	 for(y=0;y<alph;y++)taubias[y][i]+=beta[i];
	 //starting at equilibrium value
	 start[i]=invactfunc(beta[i]/alpha);
 }
 //This code assumes there are no double edges in the graph
 for(e=0;e<nedges;e++)
	 if(edgesrc[e]!=edgedest[e]){
		 for(y=0;y<alph;y++)tau[y][e]=0;
	 }else{
		 for(y=0;y<alph;y++)
			 tau[y][e]=-alpha;
	 }
}

//Version for logistic activation function
/*
	 void GLNN::initWeights()
{
 int i,e;long y;
 long double s;
 long double alpha=2;
 long double beta[n];
 long double read_epsilon;
 //Put the right global frequencies
 for(y=0;y<alph;y++)wbias[y]=logl(prfreq[y]);
 for(i=0;i<n;i++)beta[i]=1./(i+1.);
 for(i=0;i<n;i++){
	 for(y=0;y<alph;y++)w[y][i]=0;
	 read_epsilon=beta[i]*.5;
	 //alea() returns a uniform random value in [0;1]
	 for(y=0;y<alph;y++)taubias[y][i]=alea()*read_epsilon;
	 //centering
	 s=0;for(y=0;y<alph;y++)s+=taubias[y][i]*readfreq[y];
	 for(y=0;y<alph;y++)taubias[y][i]-=s;
	 //adding constant
	 for(y=0;y<alph;y++)taubias[y][i]+=beta[i];
	 //starting at equilibrium value
	 start[i]=invactfunc(beta[i]/alpha);
 }
 //This code assumes there are no double edges in the graph
 for(e=0;e<nedges;e++)
	 if(edgesrc[e]!=edgedest[e]){
		 for(y=0;y<alph;y++)tau[y][e]=0;
	 }else{
		 for(y=0;y<alph;y++)
			 tau[y][e]=-alpha;
	 }
}
*/

void GLNN::computeNextAct(long t,long x)
{
 int i,j,e;
 for(j=0;j<n;j++)
	 V[t][j]=taubias[x][j]+V[t-1][j];
 for(e=0;e<nedges;e++)
	 V[t][edgedest[e]]+=tau[x][e]*act[t-1][edgesrc[e]];
 for(i=0;i<n;i++)act[t][i]=actfunc(V[t][i]);
}

void GLNN::computeBackprop(const vector<long>& datum)
{
 int t,i,e;long x,y;
 for(i=0;i<n;i++)B[datum.size()][i]=0;
 for(t=datum.size()-1;t>=0;t--){
	 x=datum[t];
	 if(isToPredict(datum,t)){
		 for(i=0;i<n;i++)B[t][i]=w[x][i];
		 for(y=0;y<alph;y++)for(i=0;i<n;i++)
			 B[t][i]-=pred[t][y]*w[y][i];
	 }else{
		 for(i=0;i<n;i++)B[t][i]=0;
	 }
	 for(e=0;e<nedges;e++){
		 i=edgesrc[e];
		 B[t][i]+=tau[x][e]*B[t+1][edgedest[e]];
	 }
	 for(i=0;i<n;i++)B[t][i]*=deractfunc(act[t][i]);
	 for(i=0;i<n;i++)B[t][i]+=B[t+1][i];
 }
}


void GLNN::computeBPM(const vector<long>& datum)
{
 int t,i,e;long x,y;
 for(i=0;i<n;i++)metricV[datum.size()][i]=0;
 long double m,v;
 for(t=datum.size()-1;t>=0;t--){
	 for(i=0;i<n;i++){
		 if(isToPredict(datum,t)){
			 m=0;v=0;
			 for(y=0;y<alph;y++)m+=w[y][i]*pred[t][y];
			 for(y=0;y<alph;y++)v+=(w[y][i]-m)*(w[y][i]-m)*pred[t][y];
			 if(v<0){cerr<<"Neg v "<<v<<endl;v=0;}
			 metricV[t][i]=v;
		 }else metricV[t][i]=0;
		 aux[i]=0;
	 }
	 x=datum[t];
	 for(e=0;e<nedges;e++){
		 i=edgesrc[e];
		 if(i!=edgedest[e])
			 metricV[t][i]+=tau[x][e]*tau[x][e]*metricV[t+1][edgedest[e]];
		 else aux[i]+=tau[x][e];
	 }
	 for(i=0;i<n;i++){
		 metricV[t][i]*=deractfunc(act[t][i])*deractfunc(act[t][i]);
		 aux[i]*=deractfunc(act[t][i]);
		 metricV[t][i]+=(1+aux[i])*(1+aux[i])*metricV[t+1][i];
	 }
 }
}

