#include"gnn.h"
#include<eigen3/Eigen/Dense>

void GNN::Setup(long maxDatumLength)
{
 NN_SoftMaxOutput::Setup(maxDatumLength);
 B=vector<vector<long double> > (maxDatumLength+1,vector<long double>(n,0.));
 metricV=vector<vector<long double> > (maxDatumLength+1,vector<long double>(n,0.));
 tau=vector<vector<long double> >(alph,vector<long double>(nedges,0.));
 taugrad=tau;
 taubias=vector<vector<long double> > (alph,vector<long double>(n,0.));
 taubiasgrad=taubias;
 start=vector<long double> (n,0.);
 startgrad=start;
 aux=vector<long double> (n,0.);
}

void GNN::saveTauParams()
{
 prev_tau=tau;prev_taubias=taubias;prev_start=start;
}

void GNN::restoreTauParams()
{
 tau=prev_tau;taubias=prev_taubias;start=prev_start;
}

void GNN::applyTauGrad(long double rate)
{
 int e,i;long y;
 for(y=0;y<alph;y++)for(e=0;e<nedges;e++)
	 tau[y][e]+=rate*taugrad[y][e];
 for(y=0;y<alph;y++)for(i=0;i<n;i++)
	 taubias[y][i]+=rate*taubiasgrad[y][i];
 for(i=0;i<n;i++)
	 start[i]+=rate*startgrad[i];
}

int GNN::suggestedConn()
{
 switch(tauGradType){
	 case BackpropT:case FreqBackpropT:
		 return alph;
	 case BPMetricT:case UnitOutProdT:case RelL2MetricT:
		 return (int)ceil(sqrt(2.*alph));
	 default:
		 cerr<<"Need to set gradient type to use suggestedConn!"<<endl;
		 return 0;
 }
}

void GNN::initWeights()
{
 int i,e;long y;
 long double s;
 long double beta[n];
 long double read_epsilon;
 //Put the right global frequencies
 for(y=0;y<alph;y++)wbias[y]=logl(predfreq[y]);
 for(i=0;i<n;i++)beta[i]=1./(i+1.);
 for(i=0;i<n;i++){
	 for(y=0;y<alph;y++)w[y][i]=0;
	 read_epsilon=beta[i]*.5;
	 //alea() returns a uniform random value in [0;1]
	 for(y=0;y<alph;y++)taubias[y][i]=alea()*read_epsilon;
	 //centering
	 s=0;for(y=0;y<alph;y++)s+=taubias[y][i]*readfreq[y];
	 for(y=0;y<alph;y++)taubias[y][i]-=s;
	 //for logistic actfunc:
	 //adding constant so that act=1/2 is equilibrium
	 //for(y=0;y<alph;y++)taubias[y][i]+=4*(1.-beta[i])*(-1./2.);
	 //4 is the inverse derivative of actfunc at 1/2
	 start[i]=0;//starting at equilibrium value
 }
 //This code assumes there are no double edges in the graph
 for(e=0;e<nedges;e++)
	 if(edgesrc[e]!=edgedest[e]){
		 for(y=0;y<alph;y++)tau[y][e]=0;
	 }else{
		 for(y=0;y<alph;y++)
			 //tau[y][e]=4*(1.-beta[edgesrc[e]]);//for logistic
			 tau[y][e]=(1.-beta[edgesrc[e]]);
	 }

}

void GNN::setToStartAct()
{
 int i;
 for(i=0;i<n;i++)V[0][i]=start[i];
 for(i=0;i<n;i++)act[0][i]=actfunc(V[0][i]);
}

void GNN::computeNextAct(long t,long x)
{
 int i,j,e;
 for(j=0;j<n;j++)
	 V[t][j]=taubias[x][j];
 for(e=0;e<nedges;e++)
	 V[t][edgedest[e]]+=tau[x][e]*act[t-1][edgesrc[e]];
 for(i=0;i<n;i++)act[t][i]=actfunc(V[t][i]);
}

void GNN::computeBackprop(const vector<long>& datum)
{
 int t,i,e;long x,y;
 for(i=0;i<n;i++)B[datum.size()][i]=0;
 for(t=datum.size()-1;t>=0;t--){
	 x=datum[t];
	 if(isToPredict(datum,t)){
		 for(i=0;i<n;i++)B[t][i]=w[x][i];
		 for(y=0;y<alph;y++)for(i=0;i<n;i++)
			 B[t][i]-=pred[t][y]*w[y][i];
	 }else{
		 for(i=0;i<n;i++)B[t][i]=0;
	 }
	 for(e=0;e<nedges;e++){
		 i=edgesrc[e];
		 B[t][i]+=tau[x][e]*B[t+1][edgedest[e]];
	 }
	 for(i=0;i<n;i++)B[t][i]*=deractfunc(act[t][i]);
 }
}

void GNN::computeTauDiff(const vector<long>& datum)
{
 computeBackprop(datum);
 unsigned long t;
 int j,e;long x,y;

 for(y=0;y<alph;y++)for(e=0;e<nedges;e++)taugrad[y][e]=0;
 for(y=0;y<alph;y++)for(j=0;j<n;j++)taubiasgrad[y][j]=0;

 for(t=0;t<datum.size()-1;t++){
	 x=datum[t];
	 for(e=0;e<nedges;e++)
		 taugrad[x][e]+=act[t][edgesrc[e]]*B[t+1][edgedest[e]];
	 //taubiasgrad is taugrad from always-activated unit:
	 for(j=0;j<n;j++)taubiasgrad[x][j]+=B[t+1][j];
 }

 for(j=0;j<n;j++)startgrad[j]=B[0][j];

}

void GNN::computeTauGrad(const vector<long>& datum)
{
 int j,e;long y;
 long double normalize=1.l/datum.size();

 switch(tauGradType){
	 case Unset:cerr<<"Gradient type for transition parameters is not set!"<<endl;break;

	 case BackpropT:
							computeTauDiff(datum);
							for(j=0;j<n;j++)startgrad[j]*=normalize;
							for(y=0;y<alph;y++)for(e=0;e<nedges;e++)taugrad[y][e]*=normalize;
							for(y=0;y<alph;y++)for(j=0;j<n;j++)taubiasgrad[y][j]*=normalize;
							break;

	 case FreqBackpropT:
							computeTauDiff(datum);
							for(y=0;y<alph;y++)for(e=0;e<nedges;e++)taugrad[y][e]/=(datum.size()*readfreq[y]);
							for(y=0;y<alph;y++)for(j=0;j<n;j++)taubiasgrad[y][j]/=(datum.size()*readfreq[y]);
							//for(j=0;j<n;j++)startgrad[j]*=1;
							break;

	 case BPMetricT:
							computeBackprop(datum);
							computeBPM(datum);
							computeMetricInverse(datum);
							break;

	 case UnitOutProdT:
							computeBackprop(datum);
							computeUnitOutProdMetric(datum);
							computeMetricInverse(datum);
							break;

	 case RelL2MetricT:
							computeBackprop(datum);
							computeRelativeL2Metric(datum);
							computeMetricInverse(datum);
							break;

 }
}

long double square(long double x)
{
 return x*x;
}

void GNN::computeRelativeL2Metric(const vector<long>& datum)
{
 int i;unsigned long t;
 long double v,m;
 for(i=0;i<n;i++){
	 metricV[datum.size()][i]=0;
	 m=0;for(t=0;t<datum.size();t++)m+=act[t][i];
	 m/=datum.size();
	 v=0;for(t=0;t<datum.size();t++)v+=square(act[t][i]-m);
	 v/=datum.size();v+=epsilon;
	 for(t=0;t<datum.size();t++)
		 metricV[t][i]=square(deractfunc(act[t][i]))/v;
 }
}

void GNN::computeBPM(const vector<long>& datum)
{
 int t,i,e;long x,y;
 for(i=0;i<n;i++)metricV[datum.size()][i]=0;
 long double m,v;
 for(t=datum.size()-1;t>=0;t--){
	 for(i=0;i<n;i++){
		 if(isToPredict(datum,t)){
			 m=0;v=0;
			 for(y=0;y<alph;y++)m+=w[y][i]*pred[t][y];
			 for(y=0;y<alph;y++)v+=(w[y][i]-m)*(w[y][i]-m)*pred[t][y];
			 if(v<0){cerr<<"Neg v "<<v<<endl;v=0;}
			 metricV[t][i]=v;
		 }else metricV[t][i]=0;
	 }
	 x=datum[t];
	 for(e=0;e<nedges;e++){
		 i=edgesrc[e];
		 metricV[t][i]+=tau[x][e]*tau[x][e]*metricV[t+1][edgedest[e]];
	 }
	 for(i=0;i<n;i++){
		 metricV[t][i]*=deractfunc(act[t][i])*deractfunc(act[t][i]);
	 }
 }
}

void GNN::computeUnitOutProdMetric(const vector<long>& datum)
{
 int t,i;
 for(i=0;i<n;i++)metricV[datum.size()][i]=0;
 for(t=datum.size()-1;t>=0;t--)
	 for(i=0;i<n;i++)
		 metricV[t][i]=B[t][i]*B[t][i];
}

void GNN::computeMetricInverse(const vector<long>& datum)
{
 long double dampen=1.;//diagonal dampening to stabilize metric inversion. Effect is inversely proportional to data length

 unsigned long t;
 int a1,a2,j,e;long x,y;
 int np,p1,p2;
 long double norm;

 typedef Eigen::Matrix< long double , Eigen::Dynamic , Eigen::Dynamic > Mat;
 typedef Eigen::Matrix< long double , Eigen::Dynamic , 1 > Vec;
 vector<Mat> cov(alph);
 vector<Vec> grad(alph);Vec prevgrad;

 vector<long double> obj(datum.size(),0.);
 vector<long double> coeff(datum.size(),0.);
 for(j=0;j<n;j++){
	 np=1+edgesto[j].size();
	 for(y=0;y<alph;y++){cov[y].resize(np,np);cov[y].setZero();}
	 for(y=0;y<alph;y++){grad[y].resize(np);grad[y].setZero();}
	 for(t=0;t<datum.size();t++){
		 obj[t]=B[t][j];
		 coeff[t]=metricV[t][j];
	 }
	 for(t=0;t<datum.size()-1;t++){
		 x=datum[t];
		 for(p1=1;p1<np;p1++)for(p2=p1;p2<np;p2++){
			 a1=edgesto[j][p1-1].first;a2=edgesto[j][p2-1].first;
			 cov[x].coeffRef(p1,p2)+=act[t][a1]*act[t][a2]*coeff[t+1];
		 }
		 for(p2=1;p2<np;p2++){
			 a2=edgesto[j][p2-1].first;
			 cov[x].coeffRef(0,p2)+=act[t][a2]*coeff[t+1];
		 }
		 cov[x].coeffRef(0,0)+=1*coeff[t+1];
		 for(p2=1;p2<np;p2++){
			 a2=edgesto[j][p2-1].first;
			 grad[x].coeffRef(p2)+=obj[t+1]*act[t][a2];
		 }
		 grad[x].coeffRef(0)+=obj[t+1];
	 }
	 for(y=0;y<alph;y++)
		 for(p1=0;p1<np;p1++)for(p2=0;p2<p1;p2++)
			 cov[y](p1,p2)=cov[y](p2,p1);
	 for(y=0;y<alph;y++){
		 for(p1=0;p1<np;p1++)
			 cov[y](p1,p1)+=dampen;//dampen*readfreq[y];
	 }
	 for(y=0;y<alph;y++){
		 prevgrad=grad[y];
		 cov[y].ldlt().solveInPlace(grad[y]);
		 norm=prevgrad.dot(grad[y]);
		 if(norm<0){
			 cerr<<"Problem in matrix inversion, inverse is not positive semi-definite"<<endl;
#ifdef NNDEBUG
			 cerr<<"Matrix:"<<endl;
			 for(p1=0;p1<np;p1++){for(p2=0;p2<np;p2++)cerr<<cov[y](p1,p2)<<' ';cerr<<endl;}
			 cerr<<"Partial derivatives:"<<endl;
			 for(p1=0;p1<np;p1++)cerr<<prevgrad(p1)<<' ';cerr<<endl;
			 cerr<<"Gradient:"<<endl;
			 for(p1=0;p1<np;p1++)cerr<<grad[y](p1)<<' ';cerr<<endl;
#endif
		 grad[y].setZero();
		 }
	 }

	 for(y=0;y<alph;y++){
		 taubiasgrad[y][j]=grad[y](0);
		 for(p1=1;p1<np;p1++){
			 e=edgesto[j][p1-1].second;
			 taugrad[y][e]=grad[y](p1);
		 }
	 }

	 //start[i] is treated exactly as taubias['^'][i]
	 //with '^' the symbol seen before start
	 //and readfreq['^']=1./datum.size;
	 startgrad[j]=obj[0]/(coeff[0]+dampen);//+dampen/datum.size());

 }

}

