
using namespace std;

#include"hmm.h"

void HMM::Setup(long maxDataLength)
{
 PredictorNetwork::Setup(maxDataLength);
 act=vector<vector<long double> > (maxDataLength+1,vector<long double>(n,0.));
 pred=vector<vector<long double> > (maxDataLength,vector<long double>(alph,0.));
 B=vector<vector<long double> > (maxDataLength+1,vector<long double>(n,0.));
 tau=vector<long double> (nedges,0.);
 w=vector<vector<long double> > (alph,vector<long double>(n,0.));
 start=vector<long double> (n,0.);
 aux=vector<long double> (n,0.);
 gain=vector<long double> (maxDataLength+1,0);
 waux=w;tauaux=tau;
}

void HMM::saveWParams()
{
 prev_w=w;
}

void HMM::restoreWParams()
{
 w=prev_w;
}

void HMM::saveTauParams()
{
 prev_tau=tau;prev_start=start;
}

void HMM::restoreTauParams()
{
 tau=prev_tau;start=prev_start;
}

int HMM::suggestedConn()
{
 return alph>n?n:alph;
}

long double square(long double x)
{
 return x*x;
}

void HMM::initWeights()
{
 int i,e;long x;long double s;
 for(i=0;i<n;i++){
	 s=0;
	 for(x=0;x<alph;x++){
		 w[x][i]=predfreq[x]*alea();
		 s+=w[x][i];
	 }
	 for(x=0;x<alph;x++)w[x][i]/=s;
 }

//To generate a Dirichlet(1/2,1/2...): square a Gaussian r.v. and resum
//to 1
 for(i=0;i<n;i++)start[i]=square(alea_norm());
 s=0;for(i=0;i<n;i++)s+=start[i];
 for(i=0;i<n;i++)start[i]/=s;

 for(e=0;e<nedges;e++)tau[e]=square(alea_norm());
 for(i=0;i<n;i++)aux[i]=0;
 for(e=0;e<nedges;e++)aux[edgesrc[e]]+=tau[e];
 for(e=0;e<nedges;e++)tau[e]/=aux[edgesrc[e]];
}

void HMM::computePred(long t)
{
 long x;for(x=0;x<alph;x++)computePred(t,x);
}

long double HMM::computePred(long t,long x)
{
 int i;pred[t][x]=0;
 for(i=0;i<n;i++)pred[t][x]+=act[t][i]*w[x][i];
 return pred[t][x];
}

void HMM::setToStartAct()
{
 int i;
 for(i=0;i<n;i++)act[0][i]=start[i];
}

void HMM::computeNextAct(long t,long x)
{
 int i,j,e;
 for(j=0;j<n;j++)act[t][j]=0;
 for(i=0;i<n;i++)aux[i]=act[t-1][i]*w[x][i];
 for(e=0;e<nedges;e++)
	 act[t][edgedest[e]]+=tau[e]*aux[edgesrc[e]];
 long double s=0;
 for(i=0;i<n;i++)s+=act[t][i];
 if(s<=0){
	 cerr<<"Numerical problem: total mass is "<<s<<endl<<"Should switch to log-probas..."<<endl;
	 for(i=0;i<n;i++)act[t][i]=1./n;// useful when applying a trained model to validation data
 }else
	 for(i=0;i<n;i++)act[t][i]/=s;
 gain[t-1]=s;//should be equal to pred[t-1][x] up to numerical errors
}

void HMM::computeBack(const vector<long>& datum)
{
 int i,e;long x;long t;long double s;
 for(i=0;i<n;i++)B[datum.size()][i]=1;
 for(t=datum.size()-1;t!=-1;t--){
	 x=datum[t];
	 for(i=0;i<n;i++)B[t][i]=0;
	 for(e=0;e<nedges;e++)B[t][edgesrc[e]]+=tau[e]*B[t+1][edgedest[e]];//*w[x][edgesrc[e]]
	 for(i=0;i<n;i++)B[t][i]*=w[x][i];
	 s=0;
	 for(i=0;i<n;i++)s+=act[t][i]*B[t][i];
	 //s should be equal to gain[t] up to numerical errors
	 if(s<=0)
		 cerr<<"Numerical problem: total proba is "<<s<<endl<<"Should switch to log-probas..."<<endl;
	 else
		 for(i=0;i<n;i++)B[t][i]/=s;
 }
}

void HMM::applyEM(const vector<long>& datum)
{
 computeBack(datum);

 long x;int i,e;unsigned long t;
 for(x=0;x<alph;x++)for(i=0;i<n;i++)waux[x][i]=0;
 for(t=0;t<datum.size();t++){
	 x=datum[t];
	 for(i=0;i<n;i++)waux[x][i]+=act[t][i]*B[t][i];
 }

 for(e=0;e<nedges;e++)tauaux[e]=0;
 for(t=0;t<datum.size();t++){
	 x=datum[t];
	 for(e=0;e<nedges;e++)tauaux[e]+=act[t][edgesrc[e]]*w[x][edgesrc[e]]*tau[e]*B[t+1][edgedest[e]]/gain[t];
 }

 long double s;
 vector<pair<int,int> >::iterator edgesfromi;
 for(i=0;i<n;i++){
	 s=0;for(x=0;x<alph;x++)s+=waux[x][i];
	 if(s>0)for(x=0;x<alph;x++)w[x][i]=waux[x][i]/s;
 }

 for(i=0;i<n;i++)aux[i]=0;
 for(e=0;e<nedges;e++)aux[edgesrc[e]]+=tauaux[e];
 for(e=0;e<nedges;e++){
	 if(aux[edgesrc[e]]>0)tau[e]=tauaux[e]/aux[edgesrc[e]];
 }

 for(i=0;i<n;i++)aux[i]=act[0][i]*B[0][i];
 s=0;for(i=0;i<n;i++)s+=aux[i];
 if(s>0)for(i=0;i<n;i++)start[i]=aux[i]/s;

}

