#include<iostream>
#include<fstream>
#include<vector>
#include<cmath>
#include"rand.h"
#include<cfloat>
#include"indexes.h"
#include<sstream>
#include<iomanip>
#include<map>
#include<sys/resource.h>

using namespace std;

#include"hmm.h"

#ifdef VERBOSE
const bool verbose=true;
#else
const bool verbose=false;
#endif

int maxtime;int maxiter;
const long double epsilon=10*LDBL_EPSILON;//to avoid divisions by 0

//alphabet size=number of distinct symbols in data
long alph;
//raw training data and validation data
vector<unsigned char> rawdatum,rawvdatum;
//training and validation data transcribed as integers
vector<long> datum,vdatum;
//dictionary between raw data and integer representation
MyDict<unsigned char> dict;

//Reads data, prints results, etc.
#include"io.cpp"

bool isToPredict(const vector<long>&datum,long pos)
{
#ifdef XOR
 return pos>0&&(datum[pos-1]==dict.index['=']);//for the XOR problem
#else
 return true;
#endif
}

int main(int argc,char**argv)
{
 long n,conn;
 n=atoi(argv[1]);conn=atoi(argv[2]); //negative conn has special meaning, see below

#ifdef STOPBYITER
 maxiter=atoi(argv[3]);
#else
 maxtime=atoi(argv[3]);
#endif

 randomize();
 bool valid=false;if(argc>=6)valid=true;
 if(valid){
	 ifstream f2(argv[5]);getAsciiData(f2,rawvdatum,vdatum);f2.close();
 }
 ifstream f(argv[4]);getAsciiData(f,rawdatum,datum);f.close();
 unsigned long maxsize=datum.size();
 if(vdatum.size()>datum.size())maxsize=vdatum.size();

 HMM network;

 network.alph=alph;
 network.n=n;

 if(conn<=0)conn=network.suggestedConn();
 if(conn>=n)conn=n;
 if(verbose)cout<<"Using graph with "<<n<<" nodes and "<<conn<<" edges per node (including a loop for every node)."<<endl<<endl;
 network.buildErdosRenyi_loops(n,conn);

 network.Setup(maxsize);
 network.getFreqs(datum);
 network.initWeights();

 pair<long double,long>llinfo;
 long double ll,prev_ll,valid_ll=-INFINITY,validllmax=-INFINITY;
 long class_err=0,valid_class_err=0;
 int iter;

 llinfo=network.computeAct(datum);ll=llinfo.first;class_err=llinfo.second;
 if(valid){
	 llinfo=network.computeAct(vdatum,true);
	 valid_ll=llinfo.first;valid_class_err=llinfo.second;
 }
 if(verbose){
	 cout<<"Initial LL "<<ll<<"         compression  "<<-100.*ll/(8.*datum.size())<<" %        "<<endl;
	 printSample(&network,500);
 }else
	 printInfo(0,ll,valid_ll,class_err,valid_class_err);

 for(iter=0;;){
	 llinfo=network.computeAct(datum);ll=llinfo.first;
	 if(verbose)cout<<endl;

	 prev_ll=ll;network.saveParams();

	 if(verbose)cout<<"Applying EM step"<<endl;
	 network.applyEM(datum);
	 iter++;

	 llinfo=network.computeAct(datum);ll=llinfo.first;
	 if(ll<prev_ll){
		 //cerr<<"Something going wrong, direction of update does not improve likelihood"<<endl;return 1;
		 //will often happen at the last stage of convergence for numerical
		 //reasons
	 }

	 if(verbose){
		 cout<<"Sampling"<<endl;
		 printSample(&network,500);//Screws act values...
		 cout<<"LL "<<ll<<"    compression  "<<-100.*ll/(8.*datum.size())<<" %        "<<endl;
	 }

	 if(valid){
		 llinfo=network.computeAct(vdatum,true);
		 valid_ll=llinfo.first;valid_class_err=llinfo.second;
		 if(valid_ll>validllmax)validllmax=valid_ll;
		 if(verbose)cout<<"Valid LL "<<valid_ll<<"         compression  "<<-100.*valid_ll/(8.*vdatum.size())<<" %,  min  "<<-100.*validllmax/(8.*vdatum.size())<<" %        "<<endl;
	 }

	 if(!verbose)printInfo(iter,ll,valid_ll,class_err,valid_class_err);

#ifdef STOPBYITER
	 if(iter>=maxiter)return 0;
#else
	 if(elapsedtime()>maxtime)return 0;//maxtime exceeded
#endif

	 if(verbose)cout<<iter+1<<" iterations."<<endl<<endl;

 }

 if(verbose)printSample(&network,500);//Screws act values...

 return 0;
}

