#include"lnn.h"

void LNN::computeNextAct(long t,long x)
{
 int i,j,e;
 for(j=0;j<n;j++)
	 V[t][j]=V[t-1][j]+read[x][j]+bias[j];
 for(e=0;e<nedges;e++)
	 V[t][edgedest[e]]+=tau[e]*act[t-1][edgesrc[e]];
 for(i=0;i<n;i++)act[t][i]=actfunc(V[t][i]);
}

//Version closer to GLNNs
void LNN::initWeights()
{
 int i,e;long y;
 long double s;
 long double alpha=.5;
 long double gamma[n];
 long double beta[n];
 long double read_epsilon;
 //Put the right global frequencies
 for(y=0;y<alph;y++)wbias[y]=logl(predfreq[y]);
 //gamma must be strictly between alpha and 1
 for(i=0;i<n;i++)gamma[i]=1.-1./(i+2.);
 for(i=0;i<n;i++){
	 for(y=0;y<alph;y++)w[y][i]=0;
	 read_epsilon=(1.-gamma[i])/4.;
	 //alea() returns a uniform random value in [0;1]
	 for(y=0;y<alph;y++)read[y][i]=alea()*read_epsilon;
	 //centering
	 s=0;for(y=0;y<alph;y++)s+=read[y][i]*readfreq[y];
	 for(y=0;y<alph;y++)read[y][i]-=s;
	 //adding constant to get desired equilibrium value
	 bias[i]=-sqrtl(alpha*(gamma[i]+alpha-1));
	 //starting at equilibrium value
	 start[i]=invactfunc(bias[i]/alpha);
 }
 //This code assumes there are no double edges in the graph
 for(e=0;e<nedges;e++)
	 if(edgesrc[e]!=edgedest[e]){
		 tau[e]=0;
	 }else{
		 tau[e]=-alpha;
	 }
}

//Version closer to RNNs
//void LNN::initWeights()
//{
// int i,e;long y;
// long double s;
// long double read_epsilon;
// //Put the right global frequencies
// for(y=0;y<alph;y++)wbias[y]=logl(predfreq[y]);
// for(i=0;i<n;i++){
//	 for(y=0;y<alph;y++)w[y][i]=0;
//	 read_epsilon=.5;
//	 //alea() returns a uniform random value in [0;1]
//	 for(y=0;y<alph;y++)read[y][i]=alea()*read_epsilon;
//	 //centering
//	 s=0;for(y=0;y<alph;y++)s+=read[y][i]*readfreq[y];
//	 for(y=0;y<alph;y++)read[y][i]-=s;
//	 bias[i]=0;
//	 //starting at equilibrium value
//	 start[i]=0;
// }
// //This code assumes there are no double edges in the graph
// for(e=0;e<nedges;e++)
//	 if(edgesrc[e]!=edgedest[e])tau[e]=0;
//	 //else tau[e]=.5;//Memory 2
// else tau[e]=-1./(edgesrc[e]+2.);//Memory j for unit j
//}

void LNN::computeBackprop(const vector<long>& datum)
{
 int t,i,e;long x,y;
 for(i=0;i<n;i++)B[datum.size()][i]=0;
 for(t=datum.size()-1;t>=0;t--){
	 x=datum[t];
	 if(isToPredict(datum,t)){
		 for(i=0;i<n;i++)B[t][i]=w[x][i];
		 for(y=0;y<alph;y++)for(i=0;i<n;i++)
			 B[t][i]-=pred[t][y]*w[y][i];
	 }else{
		 for(i=0;i<n;i++)B[t][i]=0;
	 }
	 for(e=0;e<nedges;e++){
		 i=edgesrc[e];
		 B[t][i]+=tau[e]*B[t+1][edgedest[e]];
	 }
	 for(i=0;i<n;i++)B[t][i]*=deractfunc(act[t][i]);
	 for(i=0;i<n;i++)B[t][i]+=B[t+1][i];
 }
}


void LNN::computeBPM(const vector<long>& datum)
{
 int t,i,e;long y;
 for(i=0;i<n;i++)metricV[datum.size()][i]=0;
 long double m,v;
 for(t=datum.size()-1;t>=0;t--){
	 for(i=0;i<n;i++){
		 if(isToPredict(datum,t)){
			 m=0;v=0;
			 for(y=0;y<alph;y++)m+=w[y][i]*pred[t][y];
			 for(y=0;y<alph;y++)v+=(w[y][i]-m)*(w[y][i]-m)*pred[t][y];
			 if(v<0){cerr<<"Neg v "<<v<<endl;v=0;}
			 metricV[t][i]=v;
		 }else metricV[t][i]=0;
		 aux[i]=0;
	 }
	 for(e=0;e<nedges;e++){
		 i=edgesrc[e];
		 if(i!=edgedest[e])
			 metricV[t][i]+=tau[e]*tau[e]*metricV[t+1][edgedest[e]];
		 else aux[i]+=tau[e];
	 }
	 for(i=0;i<n;i++){
		 metricV[t][i]*=deractfunc(act[t][i])*deractfunc(act[t][i]);
		 aux[i]*=deractfunc(act[t][i]);
		 metricV[t][i]+=(1+aux[i])*(1+aux[i])*metricV[t+1][i];
	 }
 }
}

