#include<iostream>
#include<vector>
#include<cmath>
#include"rand.h"
#include<cfloat>
#include<map>

#include"network.h"
#include "rnn.h"

using namespace std;

struct LSTM: public RNN
{
 //Long Short-Term Memory. Equations following Alex Graves' thesis,
 //p38-39, with the following notation:
 //my B's are the backpropagated values
 //my act's are his b's (activities)
 //my V's are his a's (pre-activation-function variables)
 //my tau's are his w's (transmission weights)
 //my w's are the writing weights for the final softmax

 vector<vector<long double> > actinput;//input gate. actinput[t][i] is the activity of the input gate of unit i at time t
 vector<vector<long double> > Vinput;
 vector<vector<long double> > actforget;//forget gate
 vector<vector<long double> > Vforget;
 vector<vector<long double> > actoutput;//output gate
 vector<vector<long double> > Voutput;
 vector<vector<long double> > actcell;//LSTM actcell variable

 //LSTM parameters...
 //The weights for the main cell (from the input signal and from other
 //cells) are inherited from the RNN class: read[][] and tau[] and bias[]
 //Parameters for the gates
 vector<vector<long double> > readinput,readinputgrad,prev_readinput;//readinput[x][i] is the weight for the input gate of unit i when reading symbol x
 vector<long double> tauinput,tauinputgrad,prev_tauinput;//for an edge e, tauinput[e] is the connection weight from the output of the source actcell of e to the input gate of the target actcell of e
 vector<long double> wcinput,wcinputgrad,prev_wcinput;//wcinput[i] is the weight connecting actcell[i] to inputgate[i] for each actcell i
 vector<long double> biasinput,biasinputgrad,prev_biasinput;//bias for the input gate
 vector<vector<long double> > readforget,readforgetgrad,prev_readforget;//readforget[x][i] is the weight for the forget gate of unit i when reading symbol x
 vector<long double> tauforget,tauforgetgrad,prev_tauforget;//for an edge e, tauforget[e] is the connection weight from the output of the source actcell of e to the forget gate of the target actcell of e
 vector<long double> wcforget,wcforgetgrad,prev_wcforget;//wcforget[i] is the weight connecting actcell[i] to forgetgate[i] for each actcell i
 vector<long double> biasforget,biasforgetgrad,prev_biasforget;//bias for the forget gate
 vector<vector<long double> > readoutput,readoutputgrad,prev_readoutput;//readoutput[x][i] is the weight for the output gate of unit i when reading symbol x
 vector<long double> tauoutput,tauoutputgrad,prev_tauoutput;//for an edge e, tauoutput[e] is the connection weight from the output of the source actcell of e to the output gate of the target actcell of e
 vector<long double> wcoutput,wcoutputgrad,prev_wcoutput;//wcoutput[i] is the weight connecting actcell[i] to outputgate[i] for each actcell i
 vector<long double> biasoutput,biasoutputgrad,prev_biasoutput;//bias for the output gate

 vector<vector<long double> > Binput,Bforget,Boutput,Bcell;//backpropagated values wrt the gates' variables V
 //B[][] (inherited from RNN class) is backprop wrt V[][]

 long double f(long double);//act func for gates
 long double g(long double);//act func for "actcell" variable
 long double h(long double);//act func for output
 //In Graves' article "Generating sequences..." h and g are tanh while f
 //is sigmoid
 long double derf(long double);//derivative of f
 long double derg(long double);//derivative of f
 long double derh(long double);//derivative of f

 virtual void Setup(long maxDatumLength);
 virtual void saveTauParams();virtual void restoreTauParams();
 virtual void setToStartAct();
 virtual void computeNextAct(long t,long symbol_read);
 virtual void initWeights();
 virtual void computeBackprop(const vector<long>& datum);
 virtual void computeTauDiff(const vector<long>& datum);
 virtual void computeTauGrad(const vector<long>& datum);
 virtual void applyTauGrad(long double rate);

};

