#include<iostream>
#include<fstream>
#include<vector>
#include<cmath>
#include"rand.h"
#include<cfloat>
#include"indexes.h"
#include<sstream>
#include<iomanip>
#include<map>
#include<sys/resource.h>

using namespace std;

#include"rnn.h"
#include"gnn.h"
#include"lnn.h"
#include"glnn.h"
#include"lstm.h"
//#include"erfglnn.h"
//#include"nntest.h"

//#define VERBOSE
//#undef VERBOSE
//#define XOR
//#define STOPBYITER

#ifdef VERBOSE
const bool verbose=true;
#else
const bool verbose=false;
#endif

int maxtime;int maxiter;
const long double epsilon=10*LDBL_EPSILON;//to avoid divisions by 0

//alphabet size=number of distinct symbols in data
long alph;
//raw training data and validation data
vector<unsigned char> rawdatum,rawvdatum;
//training and validation data transcribed as integers
vector<long> datum,vdatum;
//dictionary between raw data and integer representation
MyDict<unsigned char> dict;

//Reads data, prints results, etc.
#include"io.cpp"

bool isToPredict(const vector<long>&datum,long pos)
{
#ifdef XOR
 return pos>0&&(datum[pos-1]==dict.index['=']);//for the XOR problem
#else
 return true;
#endif
}

int main(int argc,char**argv)
{
 long n,conn;
 if(argc<5){cerr<<"Usage: <command> <network_size> <network_connectivity> <maxtime_or_maxiter> <train_file> [valid_file]"<<endl;return -1;}
 n=atoi(argv[1]);conn=atoi(argv[2]); //negative conn has special meaning, see below

#ifdef STOPBYITER
 maxiter=atoi(argv[3]);
#else
 maxtime=atoi(argv[3]);
#endif

 randomize();
 bool valid=false;if(argc>=6)valid=true;
 if(valid){
	 ifstream f2(argv[5]);getAsciiData(f2,rawvdatum,vdatum);f2.close();
 }
 ifstream f(argv[4]);getAsciiData(f,rawdatum,datum);f.close();
 unsigned long maxsize=datum.size();
 if(vdatum.size()>datum.size())maxsize=vdatum.size();

#ifdef USE_RNN
 RNN network;
#if defined(FB)
 network.tauGradType=RNN::FreqBackpropT;
#elif defined(BPM)
 network.tauGradType=RNN::QDBPMetricT;
#elif defined(UOP)
 network.tauGradType=RNN::QDOutProdT;
#elif defined(RMS)
 network.tauGradType=RNN::RootMeanSquareT;
#endif
#endif

#ifdef USE_GNN
 GNN network;
#if defined(FB)
 network.tauGradType=GNN::FreqBackpropT;
#elif defined(BPM)
 network.tauGradType=GNN::BPMetricT;
#elif defined(UOP)
 network.tauGradType=GNN::UnitOutProdT;
#endif
#endif

#ifdef USE_LNN
 LNN network;
#if defined(FB)
 network.tauGradType=RNN::FreqBackpropT;
#elif defined(BPM)
 network.tauGradType=RNN::QDBPMetricT;
#elif defined(UOP)
 network.tauGradType=RNN::QDOutProdT;
#endif
#endif

#ifdef USE_GLNN
 GLNN network;
#if defined(FB)
 network.tauGradType=GNN::FreqBackpropT;
#elif defined(BPM)
 network.tauGradType=GNN::BPMetricT;
#elif defined(UOP)
 network.tauGradType=GNN::UnitOutProdT;
#endif
#endif

#ifdef USE_LSTM
 LSTM network;
#if defined(FB)
 network.tauGradType=RNN::FreqBackpropT;
#elif defined(BPT)
 network.tauGradType=RNN::BackpropT;
#endif
#endif

//#ifdef USE_ERFGLNN
// ErfGLNN network;
//#if defined(FB)
// network.tauGradType=ErfGNN::FreqBackpropT;
//#elif defined(BPM)
// network.tauGradType=ErfGNN::BPMetricT;
//#elif defined(UOP)
// network.tauGradType=ErfGNN::UnitOutProdT;
//#endif
//#endif

#if defined(DH)
 network.wGradType=NN_SoftMaxOutput::DiagInvHessW;
#elif defined(QDH)
 network.wGradType=NN_SoftMaxOutput::QDInvHessW;
#elif defined(BPW)
 network.wGradType=NN_SoftMaxOutput::BackpropW;
#endif

//#if defined(NNTEST)
// //TestRNN network;
// GLNN network;
// //network.tauGradType=GNN::RelL2MetricT;
// //network.tauGradType=RNN::QDBPMetricT;
// network.tauGradType=GNN::BPMetricT;
// network.wGradType=NN_SoftMaxOutput::QDInvHessWwithRenorm;
//#endif

 network.alph=alph;

 if(conn<=0)conn=network.suggestedConn();
 if(conn>=n)conn=n;
 if(verbose)cout<<"Using graph with "<<n<<" nodes and "<<conn<<" edges per node (including a loop for every node)."<<endl<<endl;
 network.buildErdosRenyi_loops(n,conn);

 network.Setup(maxsize);
 network.getFreqs(datum);
 network.initWeights();

 long double rate=1.l/(n+0.l);
 long double wrate=1.l/(n+0.l);
 //Plain backprop requires larger learning rates:
#if defined(BPW)
 wrate*=100;
#endif
#if defined(BPT)
 rate*=100;
#endif


 pair<long double,long>llinfo;
 long double ll,prev_ll,valid_ll=-INFINITY,validllmax=-INFINITY;
 long class_err=0,valid_class_err=0;
 int iter;

 llinfo=network.computeAct(datum);ll=llinfo.first;class_err=llinfo.second;
 if(valid){
	 llinfo=network.computeAct(vdatum,true);
	 valid_ll=llinfo.first;valid_class_err=llinfo.second;
 }
 if(verbose){
	 cout<<"Initial LL "<<ll<<"         compression  "<<-100.*ll/(8.*datum.size())<<" %        "<<endl;
	 printSample(&network,500);
 }else
	 printInfo(0,ll,valid_ll,class_err,valid_class_err);

 for(iter=0;;){
	 llinfo=network.computeAct(datum);ll=llinfo.first;
	 if(verbose)cout<<endl;

	 prev_ll=ll;network.saveWParams();

	 network.computeWGrad(datum);
	 iter++;

UpdateW:
	 if(verbose)cout<<"Applying w move"<<endl;
	 network.applyWGrad(wrate);

	 llinfo=network.computeAct(datum);ll=llinfo.first;
	 if(ll<prev_ll){
		 if(verbose){
			 cout<<"Cancelling w move:"<<endl;
			 cout<<"w-rate: "<<wrate<<endl;
			 cout<<"LL "<<ll<<"    compression  "<<-100.*ll/(8.*datum.size())<<" %        "<<endl;
		 }
		 wrate/=2*1.05;
		 network.restoreWParams();
		 if(wrate*datum.size()<1e-12){
			 cerr<<"Something going wrong, direction of update does not improve likelihood"<<endl;return 1;
		 }else goto UpdateW;
	 }else{
		 if(verbose){
			 cout<<"Accepting w move:"<<endl;
			 cout<<"w-rate: "<<wrate<<endl;
			 cout<<"LL "<<ll<<"         compression  "<<-100.*ll/(8.*datum.size())<<" %        "<<endl;
		 }
		 wrate*=1.1;
	 }

	 prev_ll=ll;network.saveTauParams();

	 network.computeTauGrad(datum);
	 iter++;

UpdateTrans:
	 if(verbose)cout<<"Applying trans move"<<endl;
	 network.applyTauGrad(rate);

	 llinfo=network.computeAct(datum);ll=llinfo.first;class_err=llinfo.second;
	 if(ll<prev_ll){
		 rate/=2*1.05;
		 if(verbose){
			 cout<<"Cancelling trans move:"<<endl;
			 cout<<"t-rate: "<<rate<<endl;
			 cout<<"LL "<<ll<<"    compression  "<<-100.*ll/(8.*datum.size())<<" %        "<<endl;
		 }
		 if(rate<1e-15){cerr<<"Something going wrong, direction of update does not improve likelihood"<<endl;return 1;}
		 network.restoreTauParams();
		 goto UpdateTrans;
	 }else{
		 if(verbose)cout<<"Accepting trans move."<<endl;
		 rate*=1.1;
	 }

	 if(verbose){
		 cout<<"Sampling"<<endl;
		 printSample(&network,500);//Screws act values...
		 cout<<"w-rate: "<<wrate<<endl;
		 cout<<"t-rate: "<<rate<<endl;
		 cout<<"LL "<<ll<<"    compression  "<<-100.*ll/(8.*datum.size())<<" %        "<<endl;
	 }

	 if(valid){
		 llinfo=network.computeAct(vdatum,true);
		 valid_ll=llinfo.first;valid_class_err=llinfo.second;
		 if(valid_ll>validllmax)validllmax=valid_ll;
		 if(verbose)cout<<"Valid LL "<<valid_ll<<"         compression  "<<-100.*valid_ll/(8.*vdatum.size())<<" %,  min  "<<-100.*validllmax/(8.*vdatum.size())<<" %        "<<endl;
	 }

	 if(!verbose)printInfo(iter,ll,valid_ll,class_err,valid_class_err);

#ifdef STOPBYITER
	 if(iter>=maxiter)return 0;
#else
	 if(elapsedtime()>maxtime)return 0;//maxtime exceeded
#endif

	 if(verbose)cout<<iter+1<<" iterations."<<endl<<endl;

 }

 if(verbose)printSample(&network,500);//Screws act values...

 return 0;
}

