
using namespace std;

#include"network.h"

void NN_SoftMaxOutput::fillProbFromEnergy(const vector<long double>& someenergy,vector<long double>& prob)
{
 long double max=someenergy[0],z;
 long y;
 for(y=1;y<alph;y++)if(someenergy[y]>max)max=someenergy[y];
 for(y=0;y<alph;y++)prob[y]=expl(someenergy[y]-max);
 z=0;for(y=0;y<alph;y++)z+=prob[y];
 z=1.l/z;
 for(y=0;y<alph;y++)prob[y]*=z;
}

void NN_SoftMaxOutput::computePred(long t)
{
 int i;long y;
 for(y=0;y<alph;y++)energy[y]=wbias[y];
 for(y=0;y<alph;y++)for(i=0;i<n;i++)
	 energy[y]+=act[t][i]*w[y][i];
 fillProbFromEnergy(energy,pred[t]);
}

long double NN_SoftMaxOutput::computePred(long t,long x)
{
 //For softmax output, need to compute everything
 computePred(t);return pred[t][x];
}

void NN_SoftMaxOutput::Setup(long maxDataLength)
{
 NeuralNetwork::Setup(maxDataLength);
 w=vector<vector<long double> > (alph,vector<long double>(n,0.));
 wgrad=w;
 wbias=vector<long double> (alph,0.);
 wbiasgrad=wbias;
 prev_w=w;prev_wbias=wbias;
 energy=vector<long double>(alph,0);
 hess00=vector<long double>(alph,0.);
 hess0i=vector<vector<long double> > (alph,vector<long double>(n,0.));
 hessii=hess0i;
}

void NN_SoftMaxOutput::saveWParams()
{
 prev_w=w;prev_wbias=wbias;
}

void NN_SoftMaxOutput::restoreWParams()
{
 w=prev_w;wbias=prev_wbias;
}

void NN_SoftMaxOutput::applyWGrad(long double wrate)
{
 long y;int i;
 for(y=0;y<alph;y++)for(i=0;i<n;i++)w[y][i]+=wrate*wgrad[y][i];
 for(y=0;y<alph;y++)wbias[y]+=wrate*wbiasgrad[y];
}

void NN_SoftMaxOutput::computeWDiff(const vector<long>& datum)
{
 int t,i;long x,y;
 for(x=0;x<alph;x++)for(i=0;i<n;i++)wgrad[x][i]=0;
 for(x=0;x<alph;x++)wbiasgrad[x]=0;
 for(t=datum.size()-1;t>=0;t--)if(isToPredict(datum,t)){
	 x=datum[t];
	 for(y=0;y<alph;y++)for(i=0;i<n;i++)
		 wgrad[y][i]-=pred[t][y]*act[t][i];
	 for(y=0;y<alph;y++)
		 wbiasgrad[y]-=pred[t][y];
	 for(i=0;i<n;i++)wgrad[x][i]+=act[t][i];
	 wbiasgrad[x]+=1;
 }
}

void NN_SoftMaxOutput::computeWMetric(const vector<long>& datum,bool includecrossterms)
{
 unsigned long t;
 int i;long y;long double coeff;
 for(y=0;y<alph;y++)hess00[y]=0;
 if(includecrossterms)for(y=0;y<alph;y++)for(i=0;i<n;i++)hess0i[y][i]=0;
 for(y=0;y<alph;y++)for(i=0;i<n;i++)hessii[y][i]=0;
 for(t=0;t<datum.size();t++){
	 if(isToPredict(datum,t))
		 for(y=0;y<alph;y++){
			 coeff=pred[t][y]*(1.l-pred[t][y]);
			 hess00[y]+=coeff;
			 for(i=0;i<n;i++){
				 if(includecrossterms)hess0i[y][i]+=coeff*act[t][i];
				 hessii[y][i]+=coeff*act[t][i]*act[t][i];
			 }
		 }
 }
 //Regularization
 for(y=0;y<alph;y++)hess00[y]+=predfreq[y]+epsilon;
 for(y=0;y<alph;y++)for(i=0;i<n;i++)hessii[y][i]+=predfreq[y]+epsilon;
}

//WMetric obtained from "recurrent empirical nat grad"
//{
// int i,a,b,j,e;long c,k;long double coeff;
// for(k=0;k<alph;k++)hess00[k]=0;
// for(k=0;k<alph;k++)for(a=0;a<n;a++)hess0a[k][a]=0;
// for(k=0;k<alph;k++)for(a=0;a<n;a++)hessaa[k][a]=0;
// for(i=0;i<datum.size();i++){
//   //c=datum[i];
//   if(istopredict(datum,i))
//     for(k=0;k<alph;k++){
//       coeff=-out[i][k];if(k==datum[i])coeff+=1;
//       coeff*=coeff;
//       hess00[k]+=coeff;
//       for(a=0;a<n;a++){
//         hess0a[k][a]+=coeff*lev[i][a];
//         hessaa[k][a]+=coeff*lev[i][a]*lev[i][a];
//       }
//     }
// }
// //Regularization
// for(k=0;k<alph;k++)hess00[k]+=prfreq[k]+epsilon;
// //for(k=0;k<alph;k++)for(a=0;a<n;a++)hess0a[k][a]+=freq[k];//add
// //only on diag...
// for(k=0;k<alph;k++)for(a=0;a<n;a++)hessaa[k][a]+=prfreq[k]+epsilon;
//}

long double NN_SoftMaxOutput::computeFisherMetric(const vector<long>& datum,const vector<vector<long double> >& dw,const vector<long double>& dwbias)
{
 unsigned long t;
 int i;long y;long double coeff,v,m,dE;
 long double fisherMetric=0;
 for(t=0;t<datum.size();t++){
	 if(isToPredict(datum,t)){
		 v=0;m=0;
		 for(y=0;y<alph;y++){
			 coeff=pred[t][y];
			 dE=dwbias[y];
			 for(i=0;i<n;i++)dE+=act[t][i]*dw[y][i];
			 v+=coeff*dE*dE;m+=coeff*dE;
		 }
		 v-=m*m;if(v<0)v=0;
		 fisherMetric+=v;
	 }
 }
 return fisherMetric;
}


void NN_SoftMaxOutput::computeWGrad(const vector<long>& datum)
{
 computeWDiff(datum);

 int i;long y;
 switch(wGradType){
	 case Unset:cerr<<"Gradient type for output parameters is not set!"<<endl;break;

	 case BackpropW:
							for(y=0;y<alph;y++){
								for(i=0;i<n;i++)
									wgrad[y][i]/=datum.size();
								wbiasgrad[y]/=datum.size();
							}
							break;

	 case DiagInvHessW:
							computeWMetric(datum,false);
							for(y=0;y<alph;y++){
								for(i=0;i<n;i++)
									wgrad[y][i]/=hessii[y][i];
								wbiasgrad[y]/=hess00[y];
							}
							break;

	 case QDInvHessW:
							computeWMetric(datum,true);
							for(y=0;y<alph;y++){
								for(i=0;i<n;i++)
									wgrad[y][i]=(wgrad[y][i]*hess00[y]-wbiasgrad[y]*hess0i[y][i])/(hess00[y]*hessii[y][i]-hess0i[y][i]*hess0i[y][i]);
								for(i=0;i<n;i++)wbiasgrad[y]-=hess0i[y][i]*wgrad[y][i];
								wbiasgrad[y]/=hess00[y];
							}
							break;

	 case QDInvHessWwithRenorm:
							computeWMetric(datum,true);
							vector<vector<long double> > dw=wgrad;
							vector<long double> dwbias=wbiasgrad;
							for(y=0;y<alph;y++){
								for(i=0;i<n;i++)
									dw[y][i]=(dw[y][i]*hess00[y]-dwbias[y]*hess0i[y][i])/(hess00[y]*hessii[y][i]-hess0i[y][i]*hess0i[y][i]);
								for(i=0;i<n;i++)dwbias[y]-=hess0i[y][i]*dw[y][i];
								dwbias[y]/=hess00[y];
							}
							long double alpha=0;
							for(y=0;y<alph;y++)for(i=0;i<n;i++)alpha+=dw[y][i]*wgrad[y][i];
							for(y=0;y<alph;y++)alpha+=dwbias[y]*wbiasgrad[y];
							alpha/=epsilon+computeFisherMetric(datum,dw,dwbias);
							for(y=0;y<alph;y++)for(i=0;i<n;i++)wgrad[y][i]=alpha*dw[y][i];
							for(y=0;y<alph;y++)wbiasgrad[y]=alpha*dwbias[y];
							break;

 }
}

