#include"rnn.h"

void RNN::Setup(long maxDatumLength)
{
 NN_SoftMaxOutput::Setup(maxDatumLength);
 B=vector<vector<long double> > (maxDatumLength+1,vector<long double>(n,0.));
 tau=vector<long double> (nedges,0.);
 taugrad=tau;
 bias=vector<long double>(n,0.);
 biasgrad=bias;
 read=vector<vector<long double> > (alph,vector<long double>(n,0.));
 readgrad=read;
 start=vector<long double> (n,0.);
 startgrad=start;
 metrictautau=vector<long double> (nedges,0.);
 metrictaubias=metrictautau;
 metricreadread=vector<vector<long double> > (alph,vector<long double>(n,0.));
 metricreadbias=metricreadread;
 metricbiasbias=vector<long double>(n,0.);
 metricstart=vector<long double>(n,0.);
 metricV=vector<vector<long double> > (maxDatumLength+1,vector<long double>(n,0.));
 aux=vector<long double> (n,0.);
}

void RNN::saveTauParams()
{
 prev_tau=tau;prev_bias=bias;prev_start=start;prev_read=read;
}

void RNN::restoreTauParams()
{
 tau=prev_tau;bias=prev_bias;start=prev_start;read=prev_read;
}

void RNN::applyTauGrad(long double rate)
{
 int e,i;long y;
 for(e=0;e<nedges;e++)
	 tau[e]+=rate*taugrad[e];
 for(i=0;i<n;i++)
	 bias[i]+=rate*biasgrad[i];
 for(y=0;y<alph;y++)for(i=0;i<n;i++)
	 read[y][i]+=rate*readgrad[y][i];
 for(i=0;i<n;i++)
	 start[i]+=rate*startgrad[i];
}

int RNN::suggestedConn()
{
 return alph;
}

void RNN::initWeights()
{
 int i,e;long y;
 long double s;
 long double read_epsilon;
 //Put the right global frequencies
 for(y=0;y<alph;y++)wbias[y]=logl(predfreq[y]);
 for(i=0;i<n;i++){
	 for(y=0;y<alph;y++)w[y][i]=0;
	 read_epsilon=.5;
	 //alea() returns a uniform random value in [0;1]
	 for(y=0;y<alph;y++)read[y][i]=alea()*read_epsilon;
	 //centering
	 s=0;for(y=0;y<alph;y++)s+=read[y][i]*readfreq[y];
	 for(y=0;y<alph;y++)read[y][i]-=s;
	 bias[i]=0;
	 //starting at equilibrium value
	 start[i]=0;
 }
 //This code assumes there are no double edges in the graph
 for(e=0;e<nedges;e++)
	 if(edgesrc[e]!=edgedest[e])tau[e]=0;
	 //else tau[e]=.5;//Memory 2
 else tau[e]=1.-1./(edgesrc[e]+1.);//Memory j for unit j
}

void RNN::setToStartAct()
{
 int i;
 for(i=0;i<n;i++)V[0][i]=start[i];
 for(i=0;i<n;i++)act[0][i]=actfunc(V[0][i]);
}

void RNN::computeNextAct(long t,long x)
{
 int i,j,e;
 for(j=0;j<n;j++)
	 V[t][j]=read[x][j]+bias[j];
 for(e=0;e<nedges;e++)
	 V[t][edgedest[e]]+=tau[e]*act[t-1][edgesrc[e]];
 for(i=0;i<n;i++)act[t][i]=actfunc(V[t][i]);
}

void RNN::computeBackprop(const vector<long>& datum)
{
 int t,i,e;long x,y;
 for(i=0;i<n;i++)B[datum.size()][i]=0;
 for(t=datum.size()-1;t>=0;t--){
	 x=datum[t];
	 if(isToPredict(datum,t)){
		 for(i=0;i<n;i++)B[t][i]=w[x][i];
		 for(y=0;y<alph;y++)for(i=0;i<n;i++)
			 B[t][i]-=pred[t][y]*w[y][i];
	 }else{
		 for(i=0;i<n;i++)B[t][i]=0;
	 }
	 for(e=0;e<nedges;e++){
		 i=edgesrc[e];
		 B[t][i]+=tau[e]*B[t+1][edgedest[e]];
	 }
	 for(i=0;i<n;i++)B[t][i]*=deractfunc(act[t][i]);
 }
}

void RNN::computeTauDiff(const vector<long>& datum)
{
 computeBackprop(datum);
 unsigned long t;
 int j,e;long x,y;

 //compute naive gradient
 for(e=0;e<nedges;e++)taugrad[e]=0;
 for(j=0;j<n;j++)biasgrad[j]=0;
 for(y=0;y<alph;y++)for(j=0;j<n;j++)readgrad[y][j]=0;

 for(t=0;t<datum.size()-1;t++){
	 x=datum[t];
	 for(e=0;e<nedges;e++)
		 taugrad[e]+=act[t][edgesrc[e]]*B[t+1][edgedest[e]];
	 for(j=0;j<n;j++)biasgrad[j]+=B[t+1][j];
	 for(j=0;j<n;j++)readgrad[x][j]+=B[t+1][j];
 }

 for(j=0;j<n;j++)startgrad[j]=B[0][j];
}

void RNN::computeTauGrad(const vector<long>& datum)
{
 computeTauDiff(datum);
 int j,e;long y;

 long double normalize=1.l/datum.size();

 switch(tauGradType){
	 case Unset:cerr<<"Gradient type for transition parameters is not set!"<<endl;break;

	 case BackpropT:
							for(e=0;e<nedges;e++)taugrad[e]*=normalize;
							for(j=0;j<n;j++)biasgrad[j]*=normalize;
							for(y=0;y<alph;y++)for(j=0;j<n;j++)readgrad[y][j]*=normalize;
							//treating start as read[special_start_symbol]
							for(j=0;j<n;j++)startgrad[j]*=normalize;
							break;

	 case FreqBackpropT:
							for(e=0;e<nedges;e++)taugrad[e]/=datum.size();
							for(j=0;j<n;j++)biasgrad[j]/=datum.size();
							for(y=0;y<alph;y++)for(j=0;j<n;j++)readgrad[y][j]/=(datum.size()*readfreq[y]);
							//treating start as read[special_start_symbol]
							for(j=0;j<n;j++)startgrad[j]*=1;
							break;

	 case QDOutProdT:
							computeUnitOutProdMetric(datum);
							computeQDMetric(datum);
							computeQDMetricInverse(datum);
							break;

	 case QDBPMetricT:
							computeBPM(datum);
							computeQDMetric(datum);
							computeQDMetricInverse(datum);
							break;

	 case RootMeanSquareT:
							computeUnitOutProdMetric(datum);
							computeRMSMetric(datum);
							computeQDMetricInverse(datum);//using QD inverse to compute inverse of diag matrix: slight loss of performance
							break;
 }
}

void RNN::computeUnitOutProdMetric(const vector<long>& datum)
{
 int t,i;
 for(i=0;i<n;i++)metricV[datum.size()][i]=0;
 for(t=datum.size()-1;t>=0;t--)
	 for(i=0;i<n;i++)
		 metricV[t][i]=B[t][i]*B[t][i];
}

void RNN::computeBPM(const vector<long>& datum)
{
 int t,i,e;long y;
 for(i=0;i<n;i++)metricV[datum.size()][i]=0;
 long double m,v;
 for(t=datum.size()-1;t>=0;t--){
	 for(i=0;i<n;i++){
		 if(isToPredict(datum,t)){
			 m=0;v=0;
			 for(y=0;y<alph;y++)m+=w[y][i]*pred[t][y];
			 for(y=0;y<alph;y++)v+=(w[y][i]-m)*(w[y][i]-m)*pred[t][y];
			 if(v<0){cerr<<"Neg v "<<v<<endl;v=0;}
			 metricV[t][i]=v;
		 }else metricV[t][i]=0;
	 }
	 for(e=0;e<nedges;e++)
		 metricV[t][edgesrc[e]]+=tau[e]*tau[e]*metricV[t+1][edgedest[e]];
	 for(i=0;i<n;i++)
		 metricV[t][i]*=deractfunc(act[t][i])*deractfunc(act[t][i]);
 }
}

void RNN::computeQDMetric(const vector<long>& datum)
{
//REMARK: THE FULL METRIC IS NEVER INVERTIBLE BECAUSE EXACTLY ONE INPUT
//IS ALWAYS ACTIVE, => LINEAR RELATION BETWEEN THE INPUT SIGNALS

 long double dampen=1.;//diagonal dampening to stabilize metric inversion. Effect is inversely proportional to data length

 int i,j,e;long x,y;unsigned long t;
 for(e=0;e<nedges;e++){
	 metrictautau[e]=dampen;metrictaubias[e]=0;
 }
 for(i=0;i<n;i++){
	 for(y=0;y<alph;y++){
		 metricreadread[y][i]=dampen;//dampen*readfreq[y];
		 metricreadbias[y][i]=0;
	 }
	 metricbiasbias[i]=dampen;
 }
 for(t=1;t<datum.size();t++){//at t=0, parameters not used
	 x=datum[t-1];
	 for(i=0;i<n;i++)aux[i]=metricV[t][i];
	 for(i=0;i<n;i++){
		 metricbiasbias[i]+=aux[i];
		 metricreadread[x][i]+=aux[i];
		 metricreadbias[x][i]+=aux[i];
	 }
	 for(e=0;e<nedges;e++){
		 j=edgesrc[e];i=edgedest[e];
		 metrictautau[e]+=act[t-1][j]*act[t-1][j]*aux[i];
		 metrictaubias[e]+=act[t-1][j]*aux[i];
	 }
 }

 //treating start as if reading special start symbol
 for(i=0;i<n;i++)
	 metricstart[i]=metricV[0][i]+dampen;//+dampen/datum.size();
}

void RNN::computeRMSMetric(const vector<long>& datum)
{
 long double dampen=1.;//diagonal dampening to stabilize metric inversion. Effect is inversely proportional to data length

 int i,j,e;long x,y;unsigned long t;
 for(e=0;e<nedges;e++){
	 metrictautau[e]=dampen;metrictaubias[e]=0;
 }
 for(i=0;i<n;i++){
	 for(y=0;y<alph;y++){
		 metricreadread[y][i]=dampen;//dampen*readfreq[y];
		 metricreadbias[y][i]=0;
	 }
	 metricbiasbias[i]=dampen;
 }
 for(t=1;t<datum.size();t++){//at t=0, parameters not used
	 x=datum[t-1];
	 for(i=0;i<n;i++)aux[i]=metricV[t][i];
	 for(i=0;i<n;i++){
		 metricbiasbias[i]+=aux[i];
		 metricreadread[x][i]+=aux[i];
	 }
	 for(e=0;e<nedges;e++){
		 j=edgesrc[e];i=edgedest[e];
		 metrictautau[e]+=act[t-1][j]*act[t-1][j]*aux[i];
	 }
 }

 //treating start as if reading special start symbol
 for(i=0;i<n;i++)
	 metricstart[i]=metricV[0][i]+dampen;//+dampen/datum.size();

 //Taking square root for RMS. Multiplying by datum.size() for
 //homogeneity (gradient magnitude is proportional to datum.size())
 for(e=0;e<nedges;e++)
	 metrictautau[e]=sqrtl(datum.size()*metrictautau[e]);
 for(i=0;i<n;i++){
	 metricbiasbias[i]=sqrtl(datum.size()*metricbiasbias[i]);
	 metricstart[i]=sqrtl(datum.size()*metricstart[i]);
	 for(y=0;y<alph;y++)
		 metricreadread[y][i]=sqrtl(datum.size()*metricreadread[y][i]);
 }
}

void RNN::computeQDMetricInverse(const vector<long>& datum)
{
 int i,e;long y;

 for(i=0;i<n;i++){
	 for(y=0;y<alph;y++){
		 readgrad[y][i]=(readgrad[y][i]*metricbiasbias[i]-biasgrad[i]*metricreadbias[y][i])/(metricbiasbias[i]*metricreadread[y][i]-metricreadbias[y][i]*metricreadbias[y][i]);
	 }
	 vector<pair<int,int> >::iterator edgestoi;
	 for(edgestoi=edgesto[i].begin();edgestoi!=edgesto[i].end();++edgestoi){
		 e=edgestoi->second;
		 taugrad[e]=(taugrad[e]*metricbiasbias[i]-biasgrad[i]*metrictaubias[e])/(metricbiasbias[i]*metrictautau[e]-metrictaubias[e]*metrictaubias[e]);
	 }
	 biasgrad[i]=biasgrad[i]/metricbiasbias[i];
	 for(y=0;y<alph;y++)biasgrad[i]-=readgrad[y][i]*metricreadbias[y][i];
	 for(edgestoi=edgesto[i].begin();edgestoi!=edgesto[i].end();++edgestoi){
		 e=edgestoi->second;
		 biasgrad[i]-=taugrad[e]*metrictaubias[e];
	 }
	 biasgrad[i]/=metricbiasbias[i];
 }

 for(i=0;i<n;i++)startgrad[i]/=metricstart[i];

}

