#ifndef _NEURAL_NETWORK_RNN_H
#define _NEURAL_NETWORK_RNN_H 1

#include<iostream>
#include<vector>
#include<cmath>
#include"rand.h"
#include<cfloat>
#include<map>

#include"network.h"

using namespace std;

struct RNN:public NN_SoftMaxOutput,public NN_tanhAct{
	typedef enum {Unset,BackpropT,FreqBackpropT,QDOutProdT,QDBPMetricT,RootMeanSquareT} TauGradType;
	TauGradType tauGradType;
	vector<long double> tau,taugrad,prev_tau; //tau[e]  transition weight for edge e
	vector<long double> bias,biasgrad,prev_bias; //bias[i]  bias for unit i = transition weight from always-activated unit
	vector<vector<long double> > read,readgrad,prev_read; //read[x][i] input for unit i upon reading symbol x
	vector<long double> start,prev_start,startgrad; //start[i] = value of V[i] at time 0
	vector<vector<long double> > B;  // B[t][i] backpropagated value at unit i at time t (wrt V)
	vector<vector<long double> > metricV; // metricV[t][i] metric for activity of unit i at time t, eg backpropagated metric or empirical nat grad
	vector<long double> metrictautau; // metrictautau[e] for edge e = diag entry of the metric corresponding to tau[e] 
	vector<vector<long double> > metricreadread; // metricreadread[x][i] = diag entry of the metric corresponding to read[x][i]
	vector<long double> metricbiasbias; // metricbiasbias[i] = diag entry of the metric corresponding to bias[i]
	vector<long double> metrictaubias; // metrictaubias[e] = non-diag entry of the metric between tau[e] and bias[i] with i the endpoint of edge e
	vector<vector<long double> > metricreadbias; // metricreadbias[x][i] = non-diag entry of the metric between read[x][i] and bias[i]
	vector<long double> metricstart; // = metricstart[i] = diag entry of the metric for start[i]
	vector<long double> aux; //auxiliary variable to help computations

	int suggestedConn();
	virtual void Setup(long maxDatumLength);
	virtual void saveTauParams();virtual void restoreTauParams();
	virtual void setToStartAct();
	virtual void computeNextAct(long t,long symbol_read);
	virtual void initWeights();
	virtual void computeTauGrad(const vector<long>& datum);
	virtual void applyTauGrad(long double rate);
	virtual void computeBackprop(const vector<long>& datum);
	virtual void computeTauDiff(const vector<long>& datum);
	virtual void computeUnitOutProdMetric(const vector<long>& datum);
	virtual void computeBPM(const vector<long>& datum);
	virtual void computeQDMetric(const vector<long>& datum);
	virtual void computeQDMetricInverse(const vector<long>& datum);
	virtual void computeRMSMetric(const vector<long>& datum);
};

#endif
